#version 330 core

#define INV_TAU 0.15915494392

#ifdef PCF
uniform sampler2DArrayShadow shadowMapTex;
#else
uniform sampler2DArray shadowMapTex;
#endif

uniform sampler3D envi0Tex;
uniform sampler3D envi1Tex;
uniform sampler1D cascadePlanes;
uniform sampler3D fogTex;

#if FOG_TYPE != 0
flat in float eyeFog;
#endif

layout(std140) uniform Environment {
	vec4 ambienceAndEmission;
	vec4 envi0ColorAndFactor;
	vec4 envi1Color;
	vec4 fogColor;
	mat4 fogMatrix;
	mat4 viewInv;
};

struct DirectionalLight {
	vec3 invDir;
	int shadowMapIndex;
	vec4 color;
};

struct PointLight {
	vec3 position;
	int shadowMapIndex;
	vec4 color;
	float radius;
	float attenuationFactor;
};

struct SpotLight {
	vec3 position;
	int shadowMapIndex;
	vec4 color;
	vec4 invDir;
	float depth;
	float attenuationFactor;
	float cosAngle;
	float borderWidth;
};

layout(std140) uniform Lights {
	mat4 lightMatrix[SHADOW_MAP_DEPTH];
	int numDirectionalLights;
	int numPointLights;
	int numSpotLights;
	DirectionalLight directionalLights[MAX_DIRECTIONAL_LIGHTS];
	PointLight pointLights[MAX_POINT_LIGHTS];
	SpotLight spotLights[MAX_SPOT_LIGHTS];
};

/** Returns the shadow occlusion factor for the given coordinate in shadow map space, where: 
 *   - xy: Texture coordinate 
 *   - z:  The depth value to compare
 *   - w:  The shadow map index*/
float shadowLookup(vec4 sunCoord)
{
#if defined(PCF) && defined(POISSON)
	float visibility = 0.0;
	vec4 poissonDisk[4] = vec4[](
	  vec4( -0.94201624, -0.39906216, 0.0, 0.0 ),
	  vec4( 0.94558609, -0.76890725, 0.0, 0.0 ),
	  vec4( -0.094184101, -0.92938870, 0.0, 0.0 ),
	  vec4( 0.34495938, 0.29387760, 0.0, 0.0 )
	);
	for (int i=0;i<4;i++){
	    visibility += 0.2 * texture( shadowMapTex, sunCoord.xywz + poissonDisk[i]/700.0);
	}
	return visibility + 0.2 * texture(shadowMapTex, sunCoord.xywz);
#elif defined(PCF)
	return texture(shadowMapTex, sunCoord.xywz);
#elif defined(POISSON)
	float visibility = 0.0;
	vec3 poissonDisk[4] = vec3[](
	  vec3( -0.94201624, -0.39906216, 0.0 ),
	  vec3( 0.94558609, -0.76890725, 0.0 ),
	  vec3( -0.094184101, -0.92938870, 0.0 ),
	  vec3( 0.34495938, 0.29387760, 0.0 )
	);
	for (int i=0;i<4;i++){
	    visibility += 0.2 * float(texture( shadowMapTex, sunCoord.xyw + poissonDisk[i]/700.0).r > sunCoord.z);
	}
	return visibility + 0.2 * float(texture(shadowMapTex, sunCoord.xyw).r > sunCoord.z);
#else
	return float(texture(shadowMapTex, sunCoord.xyw).r > sunCoord.z);
#endif
}

/** Returns the cube map index for the given coordinate, 
 *  using the order: -X, +X, -Y, +Y, -Z, +Z. */
int cubicToArray(vec3 co)
{
  vec3 siCo = sign(co);
  vec3 absCo = co*siCo;
  
  vec2 cmp = max(absCo.yx, absCo.zz);
  if (absCo.x>=cmp.x)
      return int(0.5*siCo.x+0.5);
  else if (absCo.y>=cmp.y)
      return int(0.5*siCo.y+2.5);
  else
      return int(0.5*siCo.z+4.5);
}

/** Returns the RGB light contribution for the given parameters: 
 *   - eyeDir: Unit vector from the camera to the surface point
 *   - lightRGB: The color of the light
 *   - matRGBA: The color of the surface point (premultiplied alpha)
 *   - s2l: Unit vector from the surface point to the light source
 *   - normal: The surface normal as a unit vector
 *   - shadow: Occlusion factor from shadowing or the spotlight cone (0 for shaded)
 *   - rgma: Reflective, glossy, metallic and ambient occlusion parameters
 *   - fresnel: Specular multiplier for this point*/
vec3 lightContrib(vec3 eyeDir, vec3 lightRGB, vec4 matRGBA, vec3 s2l, vec3 normal, float shadow, vec4 rgma, float fresnel)
{
	//Note: Tolerances are a countermeasure for shadow acne on surfaces parallel to the light

    //diffuse
    float diffuseCoefficient = max(dot(s2l, normal)*1.02 - 0.02, 0.0);
	vec3 contrib = lightRGB * matRGBA.rgb * (1.0-rgma.b) * diffuseCoefficient;
    //specular
    float gloss = 32.0*rgma.g*rgma.g;
    float specularCoefficient = fresnel * (gloss*rgma.r) * min(diffuseCoefficient*50.0, 1.0);
    //Nonzero limit required because of NV bug
    specularCoefficient *= pow(max(0.000001, dot(-eyeDir, reflect(-s2l, normal))), gloss);
	contrib += lightRGB * mix(vec3(rgma.r * matRGBA.a), matRGBA.rgb, rgma.b) * specularCoefficient;

	return shadow * contrib;
}

vec3 lightFragment(vec4 position, vec3 normal, vec4 color, vec3 emission, vec4 rgma, float fragDepth)
{
	float distance = length(position.xyz);
	vec3 eyeDir = position.xyz / distance;
	vec4 refl = vec4(reflect(eyeDir, normal), 0.0);
	refl = viewInv * refl;
	
	float fresnel = max( 0.0, (0.03 + 0.57*rgma.b + 0.12*rgma.r) + pow(normal.x*normal.x+normal.y*normal.y, 4.0)*(0.92-0.52*rgma.b-0.12*rgma.r));
	
    vec3 metalCoord = vec3(0.5+atan(refl.y, refl.x) * INV_TAU, 0.5-0.5*refl.z, 1.0-rgma.g);
    vec3 metallic = mix(envi0ColorAndFactor.rgb * texture(envi0Tex, metalCoord).rgb, envi1Color.rgb * texture(envi1Tex, metalCoord).rgb, envi0ColorAndFactor.a);
    vec3 lightColor = rgma.a * (color.rgb * (1.0-rgma.b) * ambienceAndEmission.rgb + mix(vec3(rgma.r*color.a), color.rgb, rgma.b) * metallic * fresnel);
    for(int i = 0; i<MAX_DIRECTIONAL_LIGHTS; i++)
    {
	    if(i >= numDirectionalLights)
	    	break;
	    
	    float factor;
		
		if(directionalLights[i].shadowMapIndex != -1)
		{
		    int shadowMapIndex = 0;
		    while(shadowMapIndex < CASCADE_COUNT && fragDepth > texelFetch(cascadePlanes, shadowMapIndex, 0).r)
		    	shadowMapIndex++;
		    
		    shadowMapIndex += directionalLights[i].shadowMapIndex;
		    vec4 sunCoord = lightMatrix[shadowMapIndex] * position;
		    sunCoord.w = float(shadowMapIndex);
		    factor = shadowLookup(sunCoord);
		}
		else
			factor = 1.0;
		
		lightColor += lightContrib(eyeDir, directionalLights[i].color.rgb, color, directionalLights[i].invDir, normal, factor, rgma, fresnel);
	}
	
	for(int i = 0; i<MAX_POINT_LIGHTS; i++)
    {
	    if(i >= numPointLights)
	    	break;
	    
	    vec3 s2l = pointLights[i].position - position.xyz;  
	    float dist = length(s2l);
	    s2l /= dist;
		float factor = pow(clamp(1.0-dist / pointLights[i].radius , 0.0, 1.0), pointLights[i].attenuationFactor);
		
		if(pointLights[i].shadowMapIndex != -1)
		{
		    int shadowMapIndex = cubicToArray(s2l);
			shadowMapIndex += pointLights[i].shadowMapIndex;
		    vec4 sunCoord = lightMatrix[shadowMapIndex] * position;
		    sunCoord /= sunCoord.w;
		    sunCoord.w = float(shadowMapIndex);
		    factor *= shadowLookup(sunCoord);
		}
		
		lightColor += lightContrib(eyeDir, pointLights[i].color.rgb, color, s2l, normal, factor, rgma, fresnel);
	}
	
	for(int i = 0; i<MAX_SPOT_LIGHTS; i++)
    {
	    if(i >= numSpotLights)
	    	break;
	    
	    vec3 s2l = spotLights[i].position - position.xyz;
	    float dist = length(s2l);
	    float planarDist = dot(s2l, spotLights[i].invDir.xyz);
	    s2l /= dist;
	    float factor = pow(clamp(1.0-dist / spotLights[i].depth , 0.0, 1.0), spotLights[i].attenuationFactor);
		
	    float cosA = planarDist / dist;
	    factor *= smoothstep(spotLights[i].cosAngle, spotLights[i].cosAngle+spotLights[i].borderWidth, cosA);
	    
		if(spotLights[i].shadowMapIndex != -1)
		{
		    vec4 sunCoord = lightMatrix[spotLights[i].shadowMapIndex] * position;
		    sunCoord /= sunCoord.w;
		    sunCoord.w = float(spotLights[i].shadowMapIndex);
		    factor *= shadowLookup(sunCoord);
		}
		
		lightColor += lightContrib(eyeDir, spotLights[i].color.rgb, color, s2l, normal, factor, rgma, fresnel);
	}
	
	lightColor += ambienceAndEmission.a * emission;
	
#if FOG_TYPE == 0
	return lightColor;
#else
	float fogDensity = fogColor.a * max(texture(fogTex, (fogMatrix * position).xyz).r, eyeFog);
	#if FOG_TYPE == 1
		return mix(lightColor, fogColor.rgb * color.a, min(1.0, fogDensity*distance));
	#else
		return mix(fogColor.rgb * color.a, lightColor, 1.0 / (1.0 + fogDensity*distance));
	#endif
#endif
}